from cape_privacy.spark import registry
from cape_privacy.spark.transformations import base


class MockTransformation(base.Transformation):
    identifier = "mock"

    def __init__(self, fake_arg):
        super().__init__(None)

    def __call__(self, x):
        pass


def test_get():
    registry.register(MockTransformation.identifier, MockTransformation)
    tfm_cls = registry.get("mock")
    args = {"fake_arg": 1}
    tfm_cls(**args)
    registry._registry.pop("mock")


def test_get_missing():
    tfm_cls = registry.get("plusWhat?")
    assert tfm_cls is None



from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
from pyspark import sql
from pyspark.sql import functions

from cape_privacy.spark import dtypes
from cape_privacy.spark.transformations import base
from cape_privacy.utils import typecheck

_FREQUENCY_TO_DELTA_FN = {
    "YEAR": lambda noise: pd.to_timedelta(noise * 365, unit="days"),
    "MONTH": lambda noise: pd.to_timedelta(noise * 30, unit="days"),
    "DAY": lambda noise: pd.to_timedelta(noise, unit="days"),
    "HOUR": lambda noise: pd.to_timedelta(noise, unit="hours"),
    "minutes": lambda noise: pd.to_timedelta(noise, unit="minutes"),
    "seconds": lambda noise: pd.to_timedelta(noise, unit="seconds"),
}
IntTuple = Union[int, Tuple[int, ...]]
StrTuple = Union[str, Tuple[str, ...]]


class NumericPerturbation(base.Transformation):
    """Add uniform random noise to a numeric series

    Mask a numeric series by adding uniform random noise to each value.
    The amount of noise is drawn from the interval [min, max).

    Attributes:
        dtype (dtypes.Numerics): series type
        min (int, float): the values generated will be greater or equal to min
        max (int, float): the values generated will be less than max
        seed (int), optional: a seed to initialize the random generator
    """

    identifier = "numeric-perturbation"
    type_signature = "col->col"

    def __init__(
        self,
        dtype: dtypes.DType,
        min: (int, float),
        max: (int, float),
        seed: Optional[int] = None,
    ):
        assert dtype in dtypes.Numerics
        typecheck.check_arg(min, (int, float))
        typecheck.check_arg(max, (int, float))
        typecheck.check_arg(seed, (int, type(None)))
        super().__init__(dtype)
        self._min = min
        self._max = max
        self._seed = seed

    def __call__(self, x: sql.Column):
        uniform_noise = functions.rand(seed=self._seed)
        if self._seed is not None:
            self._seed += 1
        affine_noise = self._min + uniform_noise * (self._max - self._min)
        if self._dtype is not dtypes.Double:
            affine_noise = affine_noise.astype(self._dtype)
        return x + affine_noise


class DatePerturbation(base.Transformation):
    """Add uniform random noise to a Pandas series of timestamps

    Mask a series by adding uniform random noise to the specified
    frequencies of timestamps. The amount of noise for each frequency
    is drawn from the internal [min_freq, max_freq).

    Note that seeds are currently not supported.

    Attributes:
        frequency (str, str list): one or more frequencies to perturbate
        min (int, int list): the frequency value will be greater or equal to min
        max (int, int list): the frequency value will be less than max
    """

    identifier = "date-perturbation"
    type_signature = "col->col"

    def __init__(
        self, frequency: StrTuple, min: IntTuple, max: IntTuple,
    ):
        super().__init__(dtypes.Date)
        self._frequency = _check_freq_arg(frequency)
        self._min = _check_minmax_arg(min)
        self._max = _check_minmax_arg(max)
        self._perturb_date = None

    def __call__(self, x: sql.Column):
        if self._perturb_date is None:
            self._perturb_date = self._make_perturb_udf()
        return self._perturb_date(x)

    def _make_perturb_udf(self):
        @functions.pandas_udf(dtypes.Date, functions.PandasUDFType.SCALAR)
        def perturb_date(x: pd.Series) -> pd.Series:
            rng = np.random.default_rng()
            for f, mn, mx in zip(self._frequency, self._min, self._max):
                # TODO can we switch to a lower dtype than np.int64?
                noise = rng.integers(mn, mx, size=x.shape)
                delta_fn = _FREQUENCY_TO_DELTA_FN.get(f, None)
                if delta_fn is None:
                    raise ValueError(
                        "Frequency {} must be one of {}.".format(
                            f, list(_FREQUENCY_TO_DELTA_FN.keys())
                        )
                    )
                x += delta_fn(noise)
            return x

        return perturb_date


def _check_minmax_arg(arg):
    """Checks that arg is an integer or a flat collection of integers."""
    if not isinstance(arg, (tuple, list)):
        if not isinstance(arg, int):
            raise ValueError
        return [arg]
    else:
        for a in arg:
            if not isinstance(a, int):
                raise ValueError
    return arg


def _check_freq_arg(arg):
    """Checks that arg is string or a flat collection of strings."""
    if not isinstance(arg, (tuple, list)):
        if not isinstance(arg, str):
            raise ValueError
        return [arg]
    else:
        for a in arg:
            if not isinstance(a, str):
                raise ValueError
    return arg



import pandas as pd


class PlusN:
    """A sample transform that adds n to a specific field.

    Attributes:
        field: The field that this transform will be applied to.
        n: The value to add to the field.
    """

    identifier = "plusN"
    type_signature = "col->col"

    def __init__(self, n: int = 1) -> None:
        self.n = n

    def __call__(self, column: pd.Series) -> pd.Series:
        return column + self.n



from pyspark import sql
from pyspark.sql import functions

from cape_privacy.spark.transformations import base as tfm


def transformer(transformation: tfm.Transformation, df: sql.DataFrame, field_name: str):
    field_column = functions.col(field_name)
    return df.withColumn(field_name, transformation(field_column))



import pandas as pd
import pandas.testing as pdt
from pyspark.sql import functions

from cape_privacy.spark import utils
from cape_privacy.spark.transformations import tokenizer as tkn


def _make_and_apply_tokenizer(sess, df, max_token_len, key):
    df = sess.createDataFrame(df, schema=["name"])
    tokenize = tkn.Tokenizer(max_token_len, key)
    result_df = df.select(tokenize(functions.col("name")))
    return result_df.withColumnRenamed("to_token(name)", "name").toPandas()


def test_tokenizer_simple():
    sess = utils.make_session("test.tokenizer.simple")
    test_df = pd.DataFrame({"name": ["Alice", "Bob"]})
    expected = pd.DataFrame(
        {
            "name": [
                "70a4b1a987767abf36463cd3e3f2b37144132e572fbb9b39f28bcaafe10d9b24",
                "dd4532a296deb4f114b1e7e88faefe4fb2b32c559ac15a8c6bcbdbcbc2aa4d4b",
            ]
        }
    )
    key = "secret_key"
    df = _make_and_apply_tokenizer(sess, test_df, max_token_len=None, key=key)
    pdt.assert_frame_equal(df, expected)


def test_tokenizer_is_linkable():
    sess = utils.make_session("test.tokenizer.isLinkable")
    test_df = pd.DataFrame({"name": ["Alice", "Bob"]})
    key1 = "secret_key"
    key2 = "secret_key"
    df1 = _make_and_apply_tokenizer(sess, test_df, max_token_len=None, key=key1)
    df2 = _make_and_apply_tokenizer(sess, test_df, max_token_len=None, key=key2)
    pdt.assert_frame_equal(df1, df2)


def test_tokenizer_is_not_linkable():
    sess = utils.make_session("test.tokenizer.isNotLinkable")
    test_df = pd.DataFrame({"name": ["Alice", "Bob"]})
    key1 = "secret_key"
    key2 = "not_your_secret_key"
    df1 = _make_and_apply_tokenizer(sess, test_df, max_token_len=None, key=key1)
    df2 = _make_and_apply_tokenizer(sess, test_df, max_token_len=None, key=key2)
    try:
        pdt.assert_frame_equal(df1, df2)
        raise NotImplemented  # noqa: F901
    except AssertionError:
        pass
    except NotImplemented:
        raise AssertionError


def test_tokenizer_with_max_token_len():
    sess = utils.make_session("test.tokenizer.maxTokenLen")
    test_df = pd.DataFrame({"name": ["Alice", "Bob"]})
    expected = pd.DataFrame({"name": ["70a4b1a987", "dd4532a296"]})
    max_token_len = 10
    key = "secret_key"
    df = _make_and_apply_tokenizer(sess, test_df, max_token_len=max_token_len, key=key)
    pdt.assert_frame_equal(df, expected)


def test_tokenizer_no_key():
    sess = utils.make_session("test.tokenizer.maxTokenLen")
    test_df = pd.DataFrame({"name": ["Alice", "Bob"]})
    _make_and_apply_tokenizer(sess, test_df, max_token_len=None, key=None)



import datetime

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest
import requests
import yaml

from cape_privacy import pandas as pandas_lib
from cape_privacy import spark as spark_lib
from cape_privacy.pandas.transformations import test_utils
from cape_privacy.policy import data
from cape_privacy.policy import exceptions
from cape_privacy.policy import policy as policy_lib
from cape_privacy.policy import policy_test_fixtures as fixtures


def test_parse_policy(tmp_path):
    d = tmp_path / "policy"

    d.mkdir()

    p = d / "policy.yaml"
    p.write_text(fixtures.y)

    policy = policy_lib.parse_policy(str(p.absolute()))

    assert policy.label == "test_policy"


def test_named_transform_not_found():
    pandas_lib.registry.register("plusN", test_utils.PlusN)
    d = yaml.load(
        fixtures.named_not_found_y("plusOne", "plusOneThousand", "plusN"),
        Loader=yaml.FullLoader,
    )

    df = pd.DataFrame(np.ones(5,), columns=["test"])

    p = data.Policy(**d)
    tfm = p.rules[0].transformations[0]

    with pytest.raises(exceptions.NamedTransformNotFound) as e:
        policy_lib._get_transformation(p, tfm, df, pandas_lib.dtypes)

    assert str(e.value) == (
        "Could not find transform plusOneThousand in transformations block"
    )


def test_named_transform_type_not_found():
    d = yaml.load(
        fixtures.named_not_found_y("plusOne", "plusOne", "plusM"),
        Loader=yaml.FullLoader,
    )
    p = data.Policy(**d)
    tfm = p.rules[0].transformations[0]

    with pytest.raises(exceptions.NamedTransformNotFound) as e:
        policy_lib._get_transformation(p, tfm, pandas_lib.registry, pandas_lib.dtypes)
    assert str(e.value) == "Could not find transform of type plusM in registry"


def test_parse_policy_url(httpserver):
    httpserver.expect_request("/policy").respond_with_data(fixtures.y)
    url = httpserver.url_for("/policy")
    policy = policy_lib.parse_policy(url)
    assert policy.label == "test_policy"


def test_parse_policy_invalid_url():
    with pytest.raises(requests.exceptions.ConnectionError):
        policy_lib.parse_policy("https://notapolicy.here.com/policy")


def test_parse_policy_invalid_file():
    with pytest.raises(FileNotFoundError):
        policy_lib.parse_policy("iamnotarealthingonthisfilesystem")


def test_apply_policy_pandas():
    pandas_lib.registry.register("plusN", test_utils.PlusN)
    d = yaml.load(fixtures.y, Loader=yaml.FullLoader)

    df = pd.DataFrame(np.ones(5,), columns=["test"])

    expected_df = df + 3

    p = data.Policy(**d)

    new_df = policy_lib.apply_policy(p, df)

    pdt.assert_frame_equal(new_df, expected_df)


def test_apply_complex_policies_pandas():
    d = yaml.load(fixtures.complex_y, Loader=yaml.FullLoader)

    df = pd.DataFrame(
        {
            "name": ["bob", "alice"],
            "val-int": [30, 50],
            "val-float": [32.43424, 56.64543],
            "date": [pd.Timestamp("2018-10-15"), pd.Timestamp("2016-09-10")],
        }
    )
    expected_df = pd.DataFrame(
        {
            "name": [
                "db6063546d5d6c1fd3826bc0a1d8188fa0dae1a174823eac1e8e063a073bf149",
                "4ae0639267ad49c658e8d266aa1caa51c876ed1d7ca788a0749d5189248295eb",
            ],
            "val-int": [23, 58],
            "val-float": [32.4, 56.6],
            "date": [pd.Timestamp("2018-01-01"), pd.Timestamp("2016-01-01")],
        }
    )

    p = data.Policy(**d)

    new_df = policy_lib.apply_policy(p, df)

    pdt.assert_frame_equal(new_df, expected_df)


def test_named_transformation_pandas():
    pandas_lib.registry.register("plusN", test_utils.PlusN)
    d = yaml.load(fixtures.named_y, Loader=yaml.FullLoader)

    df = pd.DataFrame(np.ones(5,), columns=["test"])

    expected_df = df + 3

    p = data.Policy(**d)

    new_df = policy_lib.apply_policy(p, df)

    pdt.assert_frame_equal(new_df, expected_df)


def test_column_redact_pandas():
    pandas_lib.registry.register("plusN", test_utils.PlusN)
    d = yaml.load(fixtures.redact_y, Loader=yaml.FullLoader)

    df = pd.DataFrame(np.ones((5, 2)), columns=["test", "apple"])

    p = data.Policy(**d)

    new_df = policy_lib.apply_policy(p, df)

    expected_df = pd.DataFrame(np.ones(5,), columns=["test"])

    expected_df = expected_df + 3

    pdt.assert_frame_equal(new_df, expected_df)


def test_apply_policy_spark():
    sess = spark_lib.utils.make_session("test.policy.applyPolicies")
    pd_df = pd.DataFrame(np.ones(5,), columns=["test"])
    expected_df = pd_df + 3
    df = sess.createDataFrame(pd_df)

    spark_lib.registry.register(test_utils.PlusN.identifier, test_utils.PlusN)
    d = yaml.load(fixtures.y, Loader=yaml.FullLoader)
    p = data.Policy(**d)
    new_df = policy_lib.apply_policy(p, df).toPandas()

    pdt.assert_frame_equal(new_df, expected_df)
    del spark_lib.registry._registry[test_utils.PlusN.identifier]


def test_apply_complex_policies_spark():
    sess = spark_lib.utils.make_session("test.policy.applyComplexPolicies")
    pd_df = pd.DataFrame(
        {
            "name": ["bob", "alice"],
            "val-int": [30, 50],
            "val-float": [32.43424, 56.64543],
            "date": [pd.Timestamp("2018-10-15"), pd.Timestamp("2016-09-10")],
        }
    )
    expected_df = pd.DataFrame(
        {
            "name": [
                "db6063546d5d6c1fd3826bc0a1d8188fa0dae1a174823eac1e8e063a073bf149",
                "4ae0639267ad49c658e8d266aa1caa51c876ed1d7ca788a0749d5189248295eb",
            ],
            "val-int": [25, 56],
            "val-float": [32.4, 56.6],
            # TODO: when these are pd.Timestamp, Spark's date_trunc is causing
            # dtype erasure. We should figure out why that's happening
            "date": [datetime.date(2018, 1, 1), datetime.date(2016, 1, 1)],
        }
    )
    df = sess.createDataFrame(pd_df)

    d = yaml.load(fixtures.complex_y, Loader=yaml.FullLoader)
    p = data.Policy(**d)
    new_df = policy_lib.apply_policy(p, df).toPandas()
    pdt.assert_frame_equal(new_df, expected_df, check_dtype=True)


def test_named_transformation_spark():
    sess = spark_lib.utils.make_session("test.policy.namedTransformations")
    pd_df = pd.DataFrame(np.ones(5,), columns=["test"])
    expected_df = pd_df + 3
    df = sess.createDataFrame(pd_df)

    spark_lib.registry.register(test_utils.PlusN.identifier, test_utils.PlusN)
    d = yaml.load(fixtures.named_y, Loader=yaml.FullLoader)
    p = data.Policy(**d)
    new_df = policy_lib.apply_policy(p, df).toPandas()

    pdt.assert_frame_equal(new_df, expected_df)
    del spark_lib.registry._registry[test_utils.PlusN.identifier]


def test_column_redaction_spark():
    sess = spark_lib.utils.make_session("test.policy.redaction")
    pd_df = pd.DataFrame(np.ones((5, 2)), columns=["test", "apple"])
    expected_df = pd.DataFrame(np.ones(5,), columns=["test"])
    expected_df = expected_df + 3
    df = sess.createDataFrame(pd_df)

    spark_lib.registry.register(test_utils.PlusN.identifier, test_utils.PlusN)
    d = yaml.load(fixtures.redact_y, Loader=yaml.FullLoader)
    p = data.Policy(**d)
    new_df = policy_lib.apply_policy(p, df).toPandas()

    pdt.assert_frame_equal(new_df, expected_df)
    del spark_lib.registry._registry[test_utils.PlusN.identifier]



import pandas as pd

from cape_privacy.pandas.transformations import base as tfm


def transformer(transformation: tfm.Transformation, df: pd.DataFrame, field_name: str):
    df[field_name] = transformation(df[field_name])
    return df



import importlib

if importlib.util.find_spec("pyspark") is None:

    def is_available():
        return False

    __all__ = ["is_available"]

else:

    from pyspark.sql import DataFrame

    from cape_privacy.spark import dtypes
    from cape_privacy.spark import registry
    from cape_privacy.spark import transformations
    from cape_privacy.spark.transformer import transformer
    from cape_privacy.spark.utils import configure_session
    from cape_privacy.spark.utils import make_session

    def is_available():
        return True

    __all__ = [
        "configure_session",
        "DataFrame",
        "dtypes",
        "is_available",
        "make_session",
        "transformations",
        "transformer",
        "registry",
    ]

del importlib



import numpy as np
import pandas as pd

String = pd.api.types.pandas_dtype(str)
Date = pd.api.types.pandas_dtype("datetime64")
# numeric types
Float = pd.api.types.pandas_dtype(np.float32)
Double = pd.api.types.pandas_dtype(np.float64)
Byte = pd.api.types.pandas_dtype(np.byte)
Short = pd.api.types.pandas_dtype(np.short)
Integer = pd.api.types.pandas_dtype(np.int32)
Long = pd.api.types.pandas_dtype(np.int64)
# groupings
Floats = (Float, Double)
Integers = (Byte, Short, Integer, Long)
Numerics = Floats + Integers



def check_arg(arg, types):
    if not isinstance(arg, types):
        if not isinstance(types, (tuple, list)):
            types = (types,)
        raise ValueError("Expected one of {}, got {}.".format(types, type(arg)))



from pyspark.sql import types

# base type
DType = types.DataType
# individual types
String = types.StringType()
Date = types.DateType()
Datetime = types.TimestampType()
# numeric types
Float = types.FloatType()
Double = types.DoubleType()
Byte = types.ByteType()
Short = types.ShortType()
Integer = types.IntegerType()
Long = types.LongType()
# groups
Floats = (Float, Double)
Integers = (Byte, Short, Integer, Long)
Numerics = Floats + Integers



from pyspark import sql
from pyspark.sql import functions

from cape_privacy.spark import dtypes
from cape_privacy.spark.transformations import base
from cape_privacy.utils import typecheck


class NumericRounding(base.Transformation):
    """Reduce the precision of a numeric series

    Round each value in the series to the given number
    of digits.

    Attributes:
        dtypes (dtypes.Numerics): series type.
        precision (int): set the number of digits.
    """

    identifier = "numeric-rounding"
    type_signature = "col->col"

    def __init__(self, dtype: dtypes.DType, precision: int):
        if dtype not in dtypes.Numerics:
            raise ValueError("NumericRounding requires a Numeric dtype.")
        typecheck.check_arg(precision, int)
        super().__init__(dtype)
        self._precision = precision

    def __call__(self, x: sql.Column):
        return functions.round(x, scale=self._precision)


class DateTruncation(base.Transformation):
    """Reduce the precision of a date series

    Truncate each date in a series to the unit (year or month)
    specified by frequency.

    Attributes:
        frequency (string): expect to be 'year' or 'month'
    """

    identifier = "date-truncation"
    type_signature = "col->col"

    def __init__(self, frequency: str):
        typecheck.check_arg(frequency, str)
        super().__init__(dtypes.Date)
        self._frequency = frequency.lower()

    def __call__(self, x: sql.Column):
        truncated = functions.date_trunc(self._frequency, x)
        return truncated.astype(self.dtype)



y = """
    label: test_policy
    version: 1
    rules:
      - match:
          name: test
        actions:
          - transform:
              type: plusN
              n: 1
          - transform:
              type: plusN
              n: 2
    """

named_y = """
    version: 1
    label: test_policy
    transformations:
      - name: plusOne
        type: plusN
        n: 1
      - name: plusTwo
        type: plusN
        n: 2
    rules:
      - match:
          name: test
        actions:
          - transform:
              name: plusOne
          - transform:
              name: plusTwo
    """


def named_not_found_y(saved_tfm, ref_tfm, tfm_type):
    return """
        label: test_policy
        version: 1
        transformations:
          - name: {saved}
            type: {type}
            n: 1
        rules:
          - match:
              name: test
            actions:
              - transform:
                  name: {ref}
    """.format(
        saved=saved_tfm, type=tfm_type, ref=ref_tfm
    )


complex_y = """
    label: test_policy
    version: 1
    rules:
      - match:
          name: val-int
        actions:
          - transform:
              type: numeric-perturbation
              dtype: Integer
              min: -10
              max: 10
              seed: 4984
      - match:
          name: val-float
        actions:
          - transform:
              type: numeric-rounding
              dtype: Double
              precision: 1
      - match:
          name: name
        actions:
          - transform:
              type: tokenizer
              key: secret_key
      - match:
          name: date
        actions:
          - transform:
              type: date-truncation
              frequency: year
    """


redact_y = """
    label: test_policy
    version: 1
    rules:
      - match:
          name: apple
        actions:
          - drop
      - match:
          name: test
        actions:
          - transform:
              type: plusN
              n: 1
          - transform:
              type: plusN
              n: 2
    """



import hashlib
import secrets

import pandas as pd
from pyspark.sql import functions

from cape_privacy.spark import dtypes
from cape_privacy.spark.transformations import base
from cape_privacy.utils import typecheck


class Tokenizer(base.Transformation):
    """Tokenizer: map a string to a token to obfuscate it.

    When applying the tokenizer to a Spark series of type string,
    each value gets mapped to a token (hexadecimal string).
    If a value is repeated several times across the series, it always
    get mapped to the same token in order to maintain the count.
    A value can be mapped to different tokens by setting the key to a
    different value.

    Attributes:
        max_token_len (int or bytes): control the token length (default
            length is 64)
        key: expect a string or byte string. if not specified, key will
            be set to a random byte string.
    """

    identifier = "tokenizer"
    type_signature = "col->col"

    def __init__(self, max_token_len=None, key=None):
        typecheck.check_arg(max_token_len, (int, type(None)))
        typecheck.check_arg(key, (str, bytes, type(None)))
        super().__init__(dtypes.String)
        self._max_token_len = max_token_len
        if isinstance(key, str):
            key = key.encode()
        self._key = key or secrets.token_bytes(8)
        self._tokenize = None

    def __call__(self, x):
        if self._tokenize is None:
            self._tokenize = self._make_tokenize_udf()
        return self._tokenize(x)

    def _make_tokenize_udf(self):
        @functions.pandas_udf(dtypes.String, functions.PandasUDFType.SCALAR)
        def to_token(x: pd.Series):
            return x.map(self._to_token)

        return to_token

    def _to_token(self, x: str):
        token = hashlib.sha256(x.encode() + self.key).hexdigest()
        if self._max_token_len is None:
            return token
        return token[: self._max_token_len]

    @property
    def key(self):
        return self._key



from typing import Callable
from typing import Dict

from cape_privacy.spark.transformations import perturbation
from cape_privacy.spark.transformations import redaction
from cape_privacy.spark.transformations import rounding
from cape_privacy.spark.transformations import tokenizer

TransformationCtor = Callable

_registry: Dict[str, TransformationCtor] = {}


def get(transformation: str) -> TransformationCtor:
    """Returns the constructor for the given key.

    Arguments:
        transformation: The key of transformation to retrieve.
    """
    return _registry.get(transformation, None)


def register(label: str, ctor: TransformationCtor):
    """Registers a new transformation constructor under the label provided.

    Arguments:
        label: The label that will be used as the key in the registry
        ctor: The transformation constructor
    """
    _registry[label] = ctor


register(perturbation.DatePerturbation.identifier, perturbation.DatePerturbation)
register(perturbation.NumericPerturbation.identifier, perturbation.NumericPerturbation)
register(rounding.NumericRounding.identifier, rounding.NumericRounding)
register(tokenizer.Tokenizer.identifier, tokenizer.Tokenizer)
register(rounding.DateTruncation.identifier, rounding.DateTruncation)
register(redaction.ColumnRedact.identifier, redaction.ColumnRedact)
register(redaction.RowRedact.identifier, redaction.RowRedact)



import numpy as np
import pandas as pd
import pandas.testing as pdt

from cape_privacy.pandas.transformations import ColumnRedact


def test_column_redact():
    redact = ColumnRedact(["b", "c"])

    df = pd.DataFrame(np.ones((5, 3)), columns=["a", "b", "c"])

    expected = pd.DataFrame(np.ones((5,)), columns=["a"])

    result = redact(df)

    pdt.assert_frame_equal(result, expected)



import datetime

import pandas as pd

from cape_privacy.pandas import dtypes
from cape_privacy.pandas.transformations import base
from cape_privacy.utils import typecheck


class NumericRounding(base.Transformation):
    """Reduce the precision of a numeric Pandas Series

    Round each value in the Pandas Series to the given number
    of digits.

    Example:
        ```
        s = pd.Series([1.384])
        round = NumericRounding(precision=1)
        round(s) # pd.Series([1.4])
        ```

    Attributes:
        dtypes (dtypes.Numerics): Pandas Series type.
        precision (int): set the number of digits.
    """

    identifier = "numeric-rounding"
    type_signature = "col->col"

    def __init__(self, dtype: dtypes.Numerics, precision: int):
        if dtype not in dtypes.Numerics:
            raise ValueError("NumericRounding requires a Numeric dtype.")
        typecheck.check_arg(precision, int)
        super().__init__(dtype)
        self._precision = precision

    def __call__(self, x: pd.Series) -> pd.Series:
        """Round each value in the Pandas Series

        Args:
            x (A Pandas Series): need to be a list of numeric values.

        Return:
            A Pandas Series with each value rounded
        """
        return self.round_numeric(x)

    def round_numeric(self, x: pd.Series):
        rounded = x.round(self._precision)
        if isinstance(rounded.dtype.type, self.dtype.type):
            return rounded
        else:
            return rounded.astype(self.dtype)


class DateTruncation(base.Transformation):
    """Reduce the precision of a date Pandas Series
    Truncate each date in a Pandas Series to the unit (year
    or month) specified by frequency.
    Example:
        ```
        s = pd.Series([pd.Timestamp("2018-10-15")])
        trunc = DateTruncation(frequency="year")
        trunc(s) # pd.Serie([pd.Timestamp("2018-01-01")])
        ```
    Attributes:
        frequency (string): expect to be 'year' or 'month'
    """

    identifier = "date-truncation"
    type_signature = "col->col"

    def __init__(self, frequency: str):
        typecheck.check_arg(frequency, str)
        super().__init__(dtypes.Date)
        self._frequency = frequency.lower()
        _check_freq_arg(self._frequency)

    def __call__(self, x: pd.Series) -> pd.Series:
        return self._trunc_date(x)

    def _trunc_date(self, x: pd.Series) -> pd.Series:
        if self._frequency == "year":
            truncated = x.values.astype("<M8[Y]")
        elif self._frequency == "month":
            truncated = x.values.astype("<M8[M]")
        elif self._frequency == "day":
            truncated = x.values.astype("<M8[D]")
        elif self._frequency == "hour":
            truncated = x.values.astype("<M8[h]")
        elif self._frequency == "minute":
            truncated = x.values.astype("<M8[m]")
        elif self._frequency == "second":
            truncated = x.values.astype("<M8[s]")
        else:
            raise ValueError

        # Use equality instead of isintance because of inheritance
        if type(x[0]) == datetime.date:
            return pd.Series(truncated).dt.date
        else:
            return pd.Series(truncated)


def _check_freq_arg(arg):
    """Checks that arg is string or a flat collection of strings."""
    freq_options = ["year", "month", "day", "hour", "minute", "second"]

    if arg not in freq_options:
        raise ValueError("Frequency {} must be one of {}.".format(arg, freq_options,))



from typing import List

import pandas as pd


class ColumnRedact:
    """Redacts columns.

    Attributes:
        columns: The columns to redact.
    """

    identifier = "column-redact"
    type_signature = "df->df"

    def __init__(self, columns: List[str]) -> None:
        self.columns = columns

    def __call__(self, df: pd.DataFrame) -> pd.DataFrame:
        return df.drop(columns=self.columns)



from typing import Callable
from typing import Dict

from cape_privacy.pandas.transformations import ColumnRedact
from cape_privacy.pandas.transformations import DatePerturbation
from cape_privacy.pandas.transformations import DateTruncation
from cape_privacy.pandas.transformations import NumericPerturbation
from cape_privacy.pandas.transformations import NumericRounding
from cape_privacy.pandas.transformations import RowRedact
from cape_privacy.pandas.transformations import Tokenizer

TransformationCtor = Callable

_registry: Dict[str, TransformationCtor] = {}


def get(transformation: str) -> TransformationCtor:
    """Returns the constructor for the given key.

    Arguments:
        transformation: The key of transformation to retrieve.
    """
    return _registry.get(transformation, None)


def register(label: str, ctor: TransformationCtor):
    """Registers a new transformation constructor under the label provided.

    Arguments:
        label: The label that will be used as the key in the registry
        ctor: The transformation constructor
    """
    _registry[label] = ctor


register(DatePerturbation.identifier, DatePerturbation)
register(NumericPerturbation.identifier, NumericPerturbation)
register(NumericRounding.identifier, NumericRounding)
register(Tokenizer.identifier, Tokenizer)
register(DateTruncation.identifier, DateTruncation)
register(ColumnRedact.identifier, ColumnRedact)
register(RowRedact.identifier, RowRedact)



import abc


class AbstractTransformation(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def dtype(self):
        pass

    @abc.abstractmethod
    def __call__(self, x):
        pass


class Transformation(AbstractTransformation):
    def __init__(self, dtype):
        self._dtype = dtype

    @property
    def dtype(self):
        return self._dtype



from typing import List

from pyspark import sql


class ColumnRedact:
    """Redacts columns from a Spark dataframe.

    Attributes:
        columns: Which columns are redacted.
    """

    identifier = "column-redact"
    type_signature = "df->df"

    def __init__(self, columns: List[str]):
        self.columns = columns

    def __call__(self, df: sql.DataFrame) -> sql.DataFrame:
        return df.drop(*self.columns)


class RowRedact:
    """Redacts rows satisfying some condition from a Spark DataFrame.

    Attributes:
        condition: When this condition evaluates to True for a row, that row
            will be dropped.
    """

    identifier = "row-redact"
    type_signature = "df->df"

    def __init__(self, condition: str):
        self.condition = condition

    def __call__(self, df: sql.DataFrame) -> sql.DataFrame:
        cond = f"NOT {self.condition}"
        return df.filter(cond)



import yaml

from .data import Policy

y = """
    label: test_policy
    transformations:
      - name: plusOne
        type: plusN
        n: 1
    rules:
      - match:
          name: test
        actions:
          - transform:
              name: plusOne
          - transform:
              type: plusN
              n: 1
      - match:
          name: test2
    """


def test_policy_class():
    d = yaml.load(y, Loader=yaml.FullLoader)

    p = Policy(**d)

    assert p.label == "test_policy"
    assert len(p.transformations) == 1

    named = p.transformations[0]
    assert named.name == "plusOne"
    assert named.type == "plusN"
    assert len(named.args) == 1

    assert named.args["n"] == 1

    rule = p.rules[0]
    assert len(p.rules) == 2
    assert len(rule.actions) == 2

    assert len(rule.transformations) == 2

    namedTransform = rule.transformations[0]
    builtinTransform = rule.transformations[1]

    assert namedTransform.field == "test"
    assert namedTransform.name == "plusOne"

    assert builtinTransform.field == "test"
    assert builtinTransform.type == "plusN"
    assert builtinTransform.args["n"] == 1



"""Contains the policy classes that are initialized from a yaml policy file.

There are five main classes with Policy being the top level class. Policy contains
the PolicySpec and NamedTransformations. PolicySpec contains Rules and Rules
contain Transformations.

    Typical usage example:

    yaml_str = "...."
    d = yaml.load(yaml_str, Loader=yaml.FullLoad)

    # **d unpacks the dictionary produced by yaml and
    # passes them in has keyword arguments.
    policy = Policy(**d)
"""

from typing import List


class Transform:
    """A actual transform that will be applied.

    Either named or function must be passed in here. The process to apply this
    transform will look at both function and named and apply the relevant one.

    Attributes:
        field: The field this transform will be applied to.
        name: The name of the named transform, referenced from
              the top level policy object.
        type: The builtin transform that will be initialized.
        kwargs: The rest of the arguments that will be passed to the transformation.
    """

    def __init__(self, field, name=None, type=None, **kwargs):
        if field == "":
            raise ValueError("Field must be specified for transformation")

        if name is None and type is None:
            raise ValueError(
                "Either named or function must be specified"
                + f" for transformation on field {field}"
            )

        if name is not None and type is not None:
            raise ValueError(
                "Both named and function cannot be "
                + "fset for transformation on field {field}"
            )

        self.field = field
        self.name = name
        self.type = type
        self.args = kwargs


class Action:
    def __init__(self, field, transform=None):
        self.transform = Transform(field, **transform)


class Rule:
    """A rule contains actionable information of a policy.

    Attributes:
        match: The match used to select a field to be transformed.
        actions: The actions to take on a matched field.
    """

    def __init__(self, match, actions=[]):
        self.actions = []
        for action in actions:
            if type(action) is dict:
                self.actions.append(Action(match["name"], **action))
            # special case for dropping a column (i.e. column redaction)
            elif type(action) is str and action == "drop":
                self.actions.append(
                    Action(
                        match["name"],
                        {"type": "column-redact", "columns": [match["name"]]},
                    )
                )

        self.transformations = [action.transform for action in self.actions]


class NamedTransform:
    """A named transformation that captures the args.

    Attributes:
        name: The name of the named transformation.
        type: The builtin type (i.e. transform) that the named transform initializes to.
        kwargs: The args that are captured by the named transform.
    """

    def __init__(self, name, type, **kwargs):
        if name == "":
            raise ValueError("Name must be specified for named transformation")

        if type == "":
            raise ValueError(f"Type must be specified for named transformation {name}")

        if len(kwargs) == 0:
            raise ValueError(
                f"Args must be specified for named transformation {self.name}"
            )

        self.name = name
        self.type = type
        self.args = kwargs


class Policy:
    """Top level policy object.

    The top level policy object holds the all of the relevant information
    for applying policy to data.

    Attributes:
        label: The label of the policy.
        version: The version of the policy.
        rules: List of rules that will be applied to a data frame.
        transformations: The named transformations for this policy.
    """

    def __init__(
        self,
        label: str,
        version: int = 1,
        rules: List[Rule] = [],
        transformations: List[NamedTransform] = [],
    ):
        self.label = label
        self.version = version

        self.transformations = [
            NamedTransform(**transform) for transform in transformations
        ]

        if len(rules) == 0:
            raise ValueError(
                f"At least one rule must be specified for policy specification {label}"
            )

        self.rules = [Rule(**rule) for rule in rules]



from cape_privacy.pandas.transformations.column_redact import ColumnRedact
from cape_privacy.pandas.transformations.perturbation import DatePerturbation
from cape_privacy.pandas.transformations.perturbation import NumericPerturbation
from cape_privacy.pandas.transformations.rounding import DateTruncation
from cape_privacy.pandas.transformations.rounding import NumericRounding
from cape_privacy.pandas.transformations.row_redact import RowRedact
from cape_privacy.pandas.transformations.tokenizer import Tokenizer

__all__ = [
    "DateTruncation",
    "DatePerturbation",
    "NumericPerturbation",
    "NumericRounding",
    "Tokenizer",
    "ColumnRedact",
    "RowRedact",
]



from cape_privacy.pandas import dtypes
from cape_privacy.pandas import registry
from cape_privacy.pandas import transformations
from cape_privacy.pandas.transformer import transformer

__all__ = ["dtypes", "transformations", "transformer", "registry"]



import numpy as np
import pandas as pd
import pandas.testing as pdt

from cape_privacy.pandas.transformations import RowRedact


def test_row_redact():
    redact = RowRedact("a > 5")

    df = pd.DataFrame(np.ones((5, 2)), columns=["a", "b"])

    df["a"].iloc[0] = 6
    df["a"].iloc[3] = 6

    expected = pd.DataFrame(np.ones((3, 2)), columns=["a", "b"], index=[1, 2, 4])

    result = redact(df)

    pdt.assert_frame_equal(result, expected)



from cape_privacy.pandas import registry
from cape_privacy.pandas.transformations import test_utils


def test_get():
    registry.register("plusN", test_utils.PlusN)
    ctor = registry.get("plusN")
    args = {"n": 1}
    ctor(**args)


def test_get_missing():
    ctor = registry.get("plusWhat?")
    assert ctor is None



"""Utils for parsing policy and applying them.

The module reads in policy as yaml and then through apply_policy
applies them to dataframes.

Example policy yaml:

    label: test_policy
    version: 1
    rules:
    - match:
        name: value
        actions:
        # Tells the policy runner to apply the transformation
        # plusN with the specified arguments.
        - transform:
            type: plusN
            n: 1
        # Tells the policy runner to apply another plusN
        # transformation.
        - transform:
            type: plusN
            n: 2

Applying policy:

    policy = parse_policy("policy.yaml")
    df = pd.DataFrame(np.ones(5,), columns=["value"])
    df = apply_policy(policy, df)
"""

import types
from typing import Callable

import pandas as pd
import requests
import validators
import yaml

from cape_privacy import pandas as pandas_lib
from cape_privacy import spark as spark_lib
from cape_privacy.policy import data
from cape_privacy.policy import exceptions


def apply_policy(policy: data.Policy, df, inplace=False):
    """Applies a Policy to some DataFrame.

    This function is responsible for inferring the type of the DataFrame, preparing the
    relevant Spark or Pandas Transformations, and applying them to produce a transformed
    DataFrame that conforms to the Policy.

    Args:
        policy: The `Policy` object that the transformed DataFrame will conform to, e.g.
            as returned by `cape_privacy.parse_policy`.
        df: The DataFrame object to transform according to `policies`. Must be of type
            pandas.DataFrame or pyspark.sql.DataFrame.
        inplace: Whether to mutate the `df` or produce a new one. This argument is only
            relevant for Pandas DataFrames, as Spark DataFrames do not support mutation.

    Raises:
        ValueError: If df is a Spark DataFrame and inplace=True, or if df is something
            other than a Pandas or Spark DataFrame.
        DependencyError: If Spark is not configured correctly in the Python environment.
        TransformNotFound, NamedTransformNotFound: If the Policy contains a reference to
            a Transformation or NamedTransformation that is unrecognized in the
            Transformation registry.
    """
    if isinstance(df, pd.DataFrame):
        registry = pandas_lib.registry
        transformer = pandas_lib.transformer
        dtypes = pandas_lib.dtypes
        if not inplace:
            result_df = df.copy()
        else:
            result_df = df
    elif not spark_lib.is_available():
        raise exceptions.DependencyError
    elif isinstance(df, spark_lib.DataFrame):
        if inplace:
            raise ValueError(
                "Spark does not support DataFrame mutation, so inplace=True is invalid."
            )
        registry = spark_lib.registry
        transformer = spark_lib.transformer
        dtypes = spark_lib.dtypes
        result_df = df
    else:
        raise ValueError(f"Expected df to be a DataFrame, found {type(df)}.")
    for rule in policy.rules:
        result_df = _do_transformations(
            policy, rule, result_df, registry, transformer, dtypes
        )
    return result_df


def parse_policy(p: str):
    """Parses a policy yaml file.

    The passed in string can either be a path to a local file or
    a URL pointing to a file. If it is a URL then requests attempts to download it.

    Args:
        p: a path string or a URL string

    Returns:
        The Policy object initialized by the yaml.
    """
    yaml_data: str

    if validators.url(p):
        yaml_data = requests.get(p).text
    else:
        with open(p) as f:
            yaml_data = f.read()

    policy = yaml.load(yaml_data, Loader=yaml.FullLoader)
    return data.Policy(**policy)


def _maybe_replace_dtype_arg(args, dtypes):
    if "dtype" in args:
        args["dtype"] = getattr(dtypes, args["dtype"])
    return args


def _get_transformation(
    policy: data.Policy, transform: data.Transform, registry: types.ModuleType, dtypes,
):
    """Looks up the correct transform class.

    If the transform is anonymous (i.e. unnamed) then it looks it up from the
    transform registry. If it is a named transform it used load_named_transform
    to find it.

    Args:
        policy: The top level policy.
        transform: The specific transform to be applied.
        registry: The module representing the transformation registry; differs for
            Spark/Pandas.
        dtypes: Passthrough; concrete dtypes to use (spark.dtypes or pandas.dtypes).

    Returns:
        The initialize transform object.

    Raises:
        TransformNotFound: The builtin transform cannot be found.
        NamedTransformNotFound: The named transform cannot be found on the
            top level policy object.
        ValueError: If neither a function or named transform exists on the transform
            arg.
    """
    if transform.type is not None:
        tfm_ctor = registry.get(transform.type)
        if tfm_ctor is None:
            raise exceptions.TransformNotFound(
                f"Could not find builtin transform {transform.type}"
            )
        tfm_args = _maybe_replace_dtype_arg(transform.args, dtypes)
        initTransform = tfm_ctor(**tfm_args)
    elif transform.name is not None:
        initTransform = _load_named_transform(policy, transform.name, registry, dtypes)
    else:
        raise ValueError(
            f"Expected type or name for transform with field {transform.field}"
        )
    return initTransform


def _do_transformations(
    policy: data.Policy,
    rule: data.Rule,
    df,
    registry: types.ModuleType,
    transformer: Callable,
    dtypes,
):
    """Applies a specific rule's transformations to a dataframe.

    For each transform, lookup the required transform class and then apply it
    to the correct column in that dataframe.

    Args:
        policy: The top level policy.
        rule: The specific rule to apply.
        df: A Pandas or Spark dataframe.
        registry: The module representing the transformation registry; differs for
            Spark/Pandas.
        transformer: A function mapping (Transformation, DataFrame, str) to a DataFrame
            that mutates a DataFrame by applying the Transformation to one of its
            columns.
        dtypes: Passthrough; concrete dtypes to use (spark.dtypes or pandas.dtypes).

    Returns:
        The transformed dataframe.
    """

    for transform in rule.transformations:
        do_transform = _get_transformation(policy, transform, registry, dtypes)
        if do_transform.type_signature == "df->df":
            df = do_transform(df)
        else:
            df = transformer(do_transform, df, transform.field)

    return df


def _load_named_transform(
    policy: data.Policy, transformLabel: str, registry: types.ModuleType, dtypes,
):
    """Attempts to load a named transform from the top level policy.

    Looks at the top level policy object for the named transform given as transformLabel
    and initializes it from the args pulled from the policy object.

    Args:
        policy: Top level policy object.
        transformLabel: The name of the named transform.
        registry: The module representing the transformation registry; differs for
            Spark/Pandas.
        dtypes: Passthrough; concrete dtypes to use (spark.dtypes or pandas.dtypes).

    Returns:
        The initialized transform object.

    Raises:
        NamedTransformNotFound: The named transform cannot be
            found in the top level policy object.
        DependencyError: If return_spark is True but PySpark is missing from the current
            environment.
    """
    found = False

    named_transforms = policy.transformations
    for transform in named_transforms:
        if transformLabel == transform.name:
            tfm_ctor = registry.get(transform.type)
            if tfm_ctor is None:
                raise exceptions.NamedTransformNotFound(
                    f"Could not find transform of type {transform.type} in registry"
                )
            tfm_args = _maybe_replace_dtype_arg(transform.args, dtypes)
            initTransform = tfm_ctor(**tfm_args)
            found = True
            break

    if not found:
        raise exceptions.NamedTransformNotFound(
            f"Could not find transform {transformLabel} in transformations block"
        )

    return initTransform



import pandas as pd
import pandas.testing as pdt

from cape_privacy.pandas.transformations import Tokenizer


def test_tokenizer():
    transform = Tokenizer(key="secret_key")

    df = pd.DataFrame({"name": ["Alice", "Bob"]})
    expected = pd.DataFrame(
        {
            "name": [
                "70a4b1a987767abf36463cd3e3f2b37144132e572fbb9b39f28bcaafe10d9b24",
                "dd4532a296deb4f114b1e7e88faefe4fb2b32c559ac15a8c6bcbdbcbc2aa4d4b",
            ]
        }
    )

    df["name"] = transform(df["name"])

    pdt.assert_frame_equal(df, expected)


def test_tokenizer_with_max_size():
    transform = Tokenizer(max_token_len=10, key="secret_key")

    df = pd.DataFrame({"name": ["Alice", "Bob"]})
    expected = pd.DataFrame({"name": ["70a4b1a987", "dd4532a296"]})

    df["name"] = transform(df["name"])

    pdt.assert_frame_equal(df, expected)



import hashlib
import secrets

import pandas as pd

from cape_privacy.pandas import dtypes
from cape_privacy.pandas.transformations import base
from cape_privacy.utils import typecheck


class Tokenizer(base.Transformation):
    """Tokenizer: map a string to a token to obfuscate it.

    When applying the Tokenizer to a Pandas Series of type string,
    each value gets mapped to a token (hexadecimal string).
    If a value is repeated several times across the series, it always
    get mapped to the same token in order to maintain the count.
    A value can be mapped to different tokens by setting the key to a
    different value.

    Example:
        ```
        s = pd.Series(['A'])
        tokenize = Tokenizer(max_token_len=5, key='secret')
        tokenize(s) # pd.Series(['40a1e'])
        ```

    Attributes:
        max_token_len (int or bytes): control the token length (default
            length is 64)
        key: expect a string or byte string. if not specified, key will
            be set to a random byte string.
    """

    identifier = "tokenizer"
    type_signature = "col->col"

    def __init__(self, max_token_len: int = None, key: str = None):
        typecheck.check_arg(max_token_len, (int, type(None)))
        typecheck.check_arg(key, (str, bytes, type(None)))
        super().__init__(dtype=dtypes.String)
        self._max_token_len = max_token_len
        if isinstance(key, str):
            key = key.encode()
        self._key = key or secrets.token_bytes(8)

    def __call__(self, series: pd.Series) -> pd.Series:
        """Map a Pandas Series to tokens.

        Args:
            series (A Pandas Series): need to be a list of strings.

        Return:
            A Pandas Series with a list of tokens represented as hexadecimal
                strings.
        """

        return series.apply(lambda x: self.to_token(x))

    def to_token(self, x):
        token = hashlib.sha256(x.encode() + self.key).hexdigest()
        if self._max_token_len is not None:
            return token[: self._max_token_len]
        else:
            return token

    @property
    def key(self):
        return self._key



import pyspark
from packaging import version
from pyspark import sql

_3_0_0_VERSION = version.Version("3.0.0")
_spark_version = version.parse(pyspark.__version__)


def configure_session(sess: sql.SparkSession, arrow=True):
    if arrow:
        if _spark_version >= _3_0_0_VERSION:
            sess.conf.set("spark.sql.execution.arrow.pyspark.enabled", "true")
        else:
            sess.conf.set("spark.sql.execution.arrow.enabled", "true")
    return sess


def make_session(name: str, arrow: bool = True):
    sess_builder = sql.SparkSession.builder
    sess_builder = sess_builder.appName(name)
    sess = sess_builder.getOrCreate()
    sess = configure_session(sess, arrow=arrow)
    return sess



import numpy as np
import pandas as pd
import pandas.testing as pdt

from cape_privacy.spark import utils
from cape_privacy.spark.transformations import redaction as rdc


def test_column_redact():
    sess = utils.make_session("test.redaction.column")
    df = pd.DataFrame(np.ones((5, 3)), columns=["a", "b", "c"])
    expected = pd.DataFrame(np.ones((5,)), columns=["a"])
    test_df = sess.createDataFrame(df, schema=["a", "b", "c"])
    redact = rdc.ColumnRedact(["b", "c"])
    result = redact(test_df).toPandas()
    pdt.assert_frame_equal(result, expected)


def test_row_redact():
    sess = utils.make_session("test.redaction.row")
    df = pd.DataFrame(np.ones((5, 2)), columns=["a", "b"])
    df["a"].iloc[0] = 6
    df["a"].iloc[3] = 6
    expected = pd.DataFrame(np.ones((3, 2)), columns=["a", "b"])
    test_df = sess.createDataFrame(df, schema=["a", "b"])
    redact = rdc.RowRedact("a > 5")
    result = redact(test_df).toPandas()
    pdt.assert_frame_equal(result, expected)



import datetime

import numpy as np
import pandas as pd
from pyspark.sql import functions

from cape_privacy.spark import dtypes
from cape_privacy.spark import utils
from cape_privacy.spark.transformations import rounding as rnd


# Utils
def _make_and_apply_rounder(sess, df, dtype, precision):
    df = sess.createDataFrame(df, schema=["data"])
    rounder = rnd.NumericRounding(dtype, precision)
    result_df = df.select(rounder(functions.col("data")))
    return result_df.toPandas()


def _make_float_data(dtype, precision=0, scale=0.1):
    data = np.arange(6, dtype=dtype).reshape((6, 1))
    delta = data * scale
    expected = np.around(data + delta, decimals=precision)
    test_df = pd.DataFrame(data + delta, columns=["data"])
    return test_df, expected


def _make_integer_data(dtype, precision):
    data = np.array([123, 1234, 12345, 123456], dtype=dtype).reshape((4, 1))
    expected = np.around(data, precision)
    test_df = pd.DataFrame(data, columns=["data"])
    return test_df, expected


def _make_date_data(sess):
    df = sess.createDataFrame([("1997-02-28",)], ["data"])
    expected = np.array(datetime.date(1997, 2, 1))
    return df, expected


def _make_datetime_data(sess):
    df = sess.createDataFrame([("1997-02-28 05:02:11",)], ["data"])
    expected = np.array(datetime.datetime(1997, 2, 1, 0, 0, 0))
    return df, expected


# Tests
def test_rounding_float():
    precision = 0
    sess = utils.make_session("test.rounding.float")
    test_df, expected = _make_float_data(np.float32, precision)
    result_df = _make_and_apply_rounder(sess, test_df, dtypes.Float, precision)
    result = result_df.values
    assert result.dtype == expected.dtype
    np.testing.assert_almost_equal(result, expected)


def test_rounding_double():
    precision = 0
    sess = utils.make_session("test.rounding.double")
    test_df, expected = _make_float_data(np.float64, precision)
    result_df = _make_and_apply_rounder(sess, test_df, dtypes.Double, precision)
    result = result_df.values
    assert result.dtype == expected.dtype
    np.testing.assert_almost_equal(result, expected)


def test_rounding_integer():
    precision = -2
    sess = utils.make_session("test.rounding.integer")
    test_df, expected = _make_integer_data(np.int32, precision)
    result_df = _make_and_apply_rounder(sess, test_df, dtypes.Integer, precision)
    result = result_df.values
    assert result.dtype == expected.dtype
    np.testing.assert_almost_equal(result, expected)


def test_rounding_long():
    precision = -2
    sess = utils.make_session("test.rounding.integer")
    test_df, expected = _make_integer_data(np.int64, precision)
    result_df = _make_and_apply_rounder(sess, test_df, dtypes.Long, precision)
    result = result_df.values
    assert result.dtype == expected.dtype
    np.testing.assert_almost_equal(result, expected)


def test_truncate_date():
    sess = utils.make_session("test.truncation.date")
    test_df, expected = _make_date_data(sess)
    truncate = rnd.DateTruncation("month")
    result_df = test_df.select(truncate(test_df.data)).toPandas()
    result = result_df.values
    assert result.dtype == expected.dtype
    np.testing.assert_equal(result, expected)



"""Installing with setuptools."""
import setuptools


with open("README.md", "r", encoding="utf8") as fh:
  long_description = fh.read()

setuptools.setup(
    name="cape-privacy",
    version="0.1.1rc0",
    packages=setuptools.find_packages(),
    python_requires=">=3.6",
    license="Apache License 2.0",
    url="https://github.com/capeprivacy/cape-python",
    description="Cape manages secure access to all of your data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Cape Privacy",
    author_email="contact@capeprivacy.com",
    install_requires=[
        "requests==2.23.0",
        "pandas==1.0.3",
        "numpy==1.18.1",
        "pyyaml==5.3.1",
        "validators==0.15.0",
    ],
    extras_require={
        "spark": ["pyspark >=2.4", "pyarrow >=0.15.1"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 3 - Alpha",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Security :: Cryptography",
    ]
)


import pandas as pd


class RowRedact:
    """Redacts rows based on the condition.

    Attributes:
        condition: The condition to be passed into the query function.
    """

    identifier = "row-redact"
    type_signature = "df->df"

    def __init__(self, condition: str) -> None:
        self.condition = condition

    def __call__(self, df: pd.DataFrame) -> pd.DataFrame:
        """Redacts rows using Dataframe.query.

        DataFrame.query returns all the fields that it matches so
        we negate it here to get the opposite.
        """

        condition = f"~({self.condition})"
        return df.query(condition)



import numpy as np
import pandas as pd
from pyspark.sql import functions

from cape_privacy.spark import dtypes
from cape_privacy.spark import utils
from cape_privacy.spark.transformations import perturbation as ptb


def _make_and_apply_numeric_ptb(sess, df, dtype, min, max):
    df = sess.createDataFrame(df, schema=["data"])
    perturb = ptb.NumericPerturbation(dtype, min=min, max=max)
    result_df = df.select(perturb(functions.col("data")))
    return result_df.toPandas()


def _make_and_apply_date_ptb(sess, df, frequency, min, max):
    df = sess.createDataFrame(df, schema=["data"])
    perturb = ptb.DatePerturbation(frequency, min, max)
    result_df = df.select(perturb(functions.col("data")))
    return result_df.withColumnRenamed("perturb_date(data)", "data").toPandas()


def test_float_ptb_bounds():
    sess = utils.make_session("test.perturbation.float.bounds")
    data = np.arange(6, dtype=np.float32).reshape((6, 1))
    test_df = pd.DataFrame(data, columns=["data"])
    lower, upper = -2, 2
    result_df = _make_and_apply_numeric_ptb(sess, test_df, dtypes.Float, lower, upper)
    result = result_df.values
    assert result.dtype == data.dtype
    noise = result - data
    lower_check = noise >= lower
    upper_check = noise <= upper
    assert lower_check.all()
    assert upper_check.all()


def test_double_ptb_bounds():
    sess = utils.make_session("test.perturbation.double.bounds")
    data = np.arange(6, dtype=np.float64).reshape((6, 1))
    test_df = pd.DataFrame(data, columns=["data"])
    lower, upper = -2, 2
    result_df = _make_and_apply_numeric_ptb(sess, test_df, dtypes.Double, lower, upper)
    result = result_df.values
    assert result.dtype == data.dtype
    noise = result - data
    lower_check = noise >= lower
    upper_check = noise <= upper
    assert lower_check.all()
    assert upper_check.all()


def test_int_ptb_bounds():
    sess = utils.make_session("test.perturbation.integer.bounds")
    data = np.arange(10, dtype=np.int32).reshape((10, 1))
    test_df = pd.DataFrame(data, columns=["data"])
    lower, upper = -3, 3
    result_df = _make_and_apply_numeric_ptb(sess, test_df, dtypes.Integer, lower, upper)
    result = result_df.values
    assert result.dtype == data.dtype
    noise = result - data
    lower_check = noise >= lower
    upper_check = noise <= upper
    assert lower_check.all()
    assert upper_check.all()


def test_byte_ptb_bounds():
    sess = utils.make_session("test.perturbation.byte.bounds")
    data = np.arange(10, dtype=np.int8).reshape((10, 1))
    test_df = pd.DataFrame(data, columns=["data"])
    lower, upper = -3, 3
    result_df = _make_and_apply_numeric_ptb(sess, test_df, dtypes.Byte, lower, upper)
    result = result_df.values
    assert result.dtype == data.dtype
    noise = result - data
    lower_check = noise >= lower
    upper_check = noise <= upper
    assert lower_check.all()
    assert upper_check.all()


def test_short_ptb_bounds():
    sess = utils.make_session("test.perturbation.short.bounds")
    data = np.arange(10, dtype=np.int16).reshape((10, 1))
    test_df = pd.DataFrame(data, columns=["data"])
    lower, upper = -3, 3
    result_df = _make_and_apply_numeric_ptb(sess, test_df, dtypes.Short, lower, upper)
    result = result_df.values
    assert result.dtype == data.dtype
    noise = result - data
    lower_check = noise >= lower
    upper_check = noise <= upper
    assert lower_check.all()
    assert upper_check.all()


def test_integer_ptb_bounds():
    sess = utils.make_session("test.perturbation.integer.bounds")
    data = np.arange(10, dtype=np.int32).reshape((10, 1))
    test_df = pd.DataFrame(data, columns=["data"])
    lower, upper = -3, 3
    result_df = _make_and_apply_numeric_ptb(sess, test_df, dtypes.Integer, lower, upper)
    result = result_df.values
    assert result.dtype == data.dtype
    noise = result - data
    lower_check = noise >= lower
    upper_check = noise <= upper
    assert lower_check.all()
    assert upper_check.all()


def test_long_ptb_bounds():
    sess = utils.make_session("test.perturbation.long.bounds")
    data = np.arange(10, dtype=np.int64).reshape((10, 1))
    test_df = pd.DataFrame(data, columns=["data"])
    lower, upper = -3, 3
    result_df = _make_and_apply_numeric_ptb(sess, test_df, dtypes.Long, lower, upper)
    result = result_df.values
    assert result.dtype == data.dtype
    noise = result - data
    lower_check = noise >= lower
    upper_check = noise <= upper
    assert lower_check.all()
    assert upper_check.all()


def test_date_perturbation_singlefreq_bounds():
    sess = utils.make_session("test.perturbation.date.bounds.singleFrequency")
    data = pd.to_datetime(["1997-03-15", "2020-06-24"])
    test_df = pd.DataFrame(data, columns=["data"])
    frequencies = ["YEAR", "MONTH", "DAY"]
    num_days = [365, 30, 1]
    lower, upper = -2, 2
    for freq, days in zip(frequencies, num_days):
        result_df = _make_and_apply_date_ptb(sess, test_df, freq, lower, upper)
        result_df = result_df.apply(pd.to_datetime)
        noise_df = result_df - test_df
        lower_check = noise_df >= pd.to_timedelta(lower * days, unit="days")
        upper_check = noise_df <= pd.to_timedelta(upper * days, unit="days")
        assert lower_check.values.all()
        assert upper_check.values.all()


def test_date_perturbation_multifreq_bounds():
    sess = utils.make_session("test.perturbation.date.bounds.singleFrequency")
    data = pd.to_datetime(["1997-03-15", "2020-06-24"])
    test_df = pd.DataFrame(data, columns=["data"])
    frequency = ("MONTH", "DAY")
    lower, upper = (-1, -30), (1, 30)
    result_df = _make_and_apply_date_ptb(sess, test_df, frequency, lower, upper)
    result_df = result_df.apply(pd.to_datetime)
    noise_df = result_df - test_df
    lower_check = noise_df >= pd.to_timedelta(-60, unit="days")
    upper_check = noise_df <= pd.to_timedelta(60, unit="days")
    assert lower_check.values.all()
    assert upper_check.values.all()



from cape_privacy.spark.transformations.perturbation import DatePerturbation
from cape_privacy.spark.transformations.perturbation import NumericPerturbation
from cape_privacy.spark.transformations.redaction import ColumnRedact
from cape_privacy.spark.transformations.redaction import RowRedact
from cape_privacy.spark.transformations.rounding import DateTruncation
from cape_privacy.spark.transformations.rounding import NumericRounding
from cape_privacy.spark.transformations.tokenizer import Tokenizer

__all__ = [
    "DatePerturbation",
    "NumericPerturbation",
    "DateTruncation",
    "NumericRounding",
    "Tokenizer",
    "ColumnRedact",
    "RowRedact",
]



from cape_privacy.policy.data import Policy
from cape_privacy.policy.exceptions import NamedTransformNotFound
from cape_privacy.policy.exceptions import TransformNotFound
from cape_privacy.policy.policy import parse_policy

__all__ = [
    "parse_policy",
    "Policy",
    "NamedTransformNotFound",
    "TransformNotFound",
]



import datetime

import numpy as np
import pandas as pd
import pandas.testing as pdt

from cape_privacy.pandas import dtypes
from cape_privacy.pandas.transformations import DateTruncation
from cape_privacy.pandas.transformations import NumericRounding


def _make_apply_numeric_rounding(input, expected_output, ctype, dtype):
    transform = NumericRounding(dtype=ctype, precision=1)
    df = pd.DataFrame({"amount": input}).astype(dtype)
    expected = pd.DataFrame({"amount": expected_output}).astype(dtype)
    df["amount"] = transform(df.amount)
    return df, expected


def _make_apply_datetruncation(frequency, input_date, expected_date):
    transform = DateTruncation(frequency=frequency)
    df = pd.DataFrame({"date": [input_date]})
    expected = pd.DataFrame({"date": [expected_date]})
    df["date"] = transform(df.date)
    return df, expected


def test_rounding_float32():
    input = [10.8834, 4.21221]
    expected_output = [10.9, 4.2]
    df, expected = _make_apply_numeric_rounding(
        input, expected_output, dtypes.Float, np.float32
    )
    pdt.assert_frame_equal(df, expected)


def test_rounding_float64():
    input = [10.8834, 4.21221]
    expected_output = [10.9, 4.2]
    df, expected = _make_apply_numeric_rounding(
        input, expected_output, dtypes.Double, np.float64
    )
    pdt.assert_frame_equal(df, expected)


def test_truncate_date_year():
    input_date = datetime.date(year=2018, month=10, day=3)
    expected_date = datetime.date(year=2018, month=1, day=1)
    df, expected = _make_apply_datetruncation("YEAR", input_date, expected_date)
    pdt.assert_frame_equal(df, expected)


def test_truncate_datetime_year():
    input_date = pd.Timestamp(year=2018, month=10, day=3)
    expected_date = pd.Timestamp(year=2018, month=1, day=1)
    df, expected = _make_apply_datetruncation("YEAR", input_date, expected_date)
    pdt.assert_frame_equal(df, expected)


def test_truncate_datetime_month():
    input_date = pd.Timestamp(year=2018, month=10, day=3, hour=9, minute=20, second=25)
    expected_date = pd.Timestamp(year=2018, month=10, day=1, hour=0, minute=0, second=0)
    df, expected = _make_apply_datetruncation("MONTH", input_date, expected_date)
    pdt.assert_frame_equal(df, expected)


def test_truncate_datetime_day():
    input_date = pd.Timestamp(year=2018, month=10, day=3, hour=9, minute=20, second=25)
    expected_date = pd.Timestamp(year=2018, month=10, day=3, hour=0, minute=0, second=0)
    df, expected = _make_apply_datetruncation("DAY", input_date, expected_date)
    pdt.assert_frame_equal(df, expected)


def test_truncate_datetime_hour():
    input_date = pd.Timestamp(year=2018, month=10, day=3, hour=9, minute=20, second=25)
    expected_date = pd.Timestamp(year=2018, month=10, day=3, hour=9, minute=0, second=0)
    df, expected = _make_apply_datetruncation("hour", input_date, expected_date)
    pdt.assert_frame_equal(df, expected)


def test_truncate_datetime_minute():
    input_date = datetime.datetime(
        year=2018, month=10, day=3, hour=9, minute=20, second=25
    )
    expected_date = datetime.datetime(
        year=2018, month=10, day=3, hour=9, minute=20, second=0
    )
    df, expected = _make_apply_datetruncation("minute", input_date, expected_date)
    pdt.assert_frame_equal(df, expected)


def test_truncate_datetime_second():
    input_date = pd.Timestamp(year=2018, month=10, day=3, hour=9, minute=20, second=25)
    expected_date = pd.Timestamp(
        year=2018, month=10, day=3, hour=9, minute=20, second=25
    )
    df, expected = _make_apply_datetruncation("second", input_date, expected_date)
    pdt.assert_frame_equal(df, expected)



import datetime
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd

from cape_privacy.pandas import dtypes
from cape_privacy.pandas.transformations import base
from cape_privacy.utils import typecheck

_FREQUENCY_TO_DELTA_FN = {
    "YEAR": lambda noise: pd.to_timedelta(noise * 365, unit="days"),
    "MONTH": lambda noise: pd.to_timedelta(noise * 30, unit="days"),
    "DAY": lambda noise: pd.to_timedelta(noise, unit="days"),
    "HOUR": lambda noise: pd.to_timedelta(noise, unit="hours"),
    "minutes": lambda noise: pd.to_timedelta(noise, unit="minutes"),
    "seconds": lambda noise: pd.to_timedelta(noise, unit="seconds"),
}
IntTuple = Union[int, Tuple[int, ...]]
StrTuple = Union[str, Tuple[str, ...]]


class NumericPerturbation(base.Transformation):
    """Add uniform random noise to a numeric Pandas series

    Mask a numeric Pandas series by adding uniform random
    noise to each value. The amount of noise is drawn from
    the interval [min, max).

    Example:
        ```
        s = pd.Series([0, 1, 2, 3, 4])
        perturb = NumericPerturbation(dtype=Integer, min=-10, max=10, seed=123)
        perturb(s) # pd.Series([3, -7, -3, -3])
        ```

    Attributes:
        dtype (dtypes.Numerics): Pandas Series type
        min (int, float): the values generated will be greater then or equal to min
        max (int, float): the values generated will be less than max
        seed (int), optional: a seed to initialize the random generator
    """

    identifier = "numeric-perturbation"
    type_signature = "col->col"

    def __init__(
        self,
        dtype: dtypes.Numerics,
        min: Union[int, float],
        max: Union[int, float],
        seed: Optional[int] = None,
    ):
        assert dtype in dtypes.Numerics
        typecheck.check_arg(min, (int, float))
        typecheck.check_arg(max, (int, float))
        typecheck.check_arg(seed, (int, type(None)))
        super().__init__(dtype)
        self._min = min
        self._max = max
        self._rng = np.random.default_rng(seed=seed)

    def __call__(self, x: pd.Series) -> pd.Series:
        noise = pd.Series(self._rng.uniform(self._min, self._max, size=x.shape))
        if not isinstance(noise.dtype.type, self.dtype.type):
            noise = noise.astype(self.dtype)
        return x + noise


class DatePerturbation(base.Transformation):
    """Add uniform random noise to a Pandas series of timestamps

    Mask a Pandas series by adding uniform random noise to the
    specified frequencies of timestamps. The amount of noise for
    each frequency is drawn from the internal [min_freq, max_freq).

    Example:
        ```
        s = pd.Series([datetime.date(year=2020, month=2, day=15)])
        perturb = DatePerturbation(frequency="MONTH", min=-10, max=10, seed=1234)
        perturb(s) # pd.Series([datetime.date(year=2020, month=11, day=11)])
        ```

    Attributes:
        frequency (str, str list): one or more frequencies to perturbate
        min (int, int list): the frequency value will be greater or equal to min
        max (int, int list): the frequency value will be less than max
        seed (int), optional: a seed to initialize the random generator
    """

    identifier = "date-perturbation"
    type_signature = "col->col"

    def __init__(
        self,
        frequency: StrTuple,
        min: IntTuple,
        max: IntTuple,
        seed: Optional[int] = None,
    ):
        super().__init__(dtypes.Date)
        self._frequency = _check_freq_arg(frequency)
        self._min = _check_minmax_arg(min)
        self._max = _check_minmax_arg(max)
        self._rng = np.random.default_rng(seed)

    def __call__(self, x: pd.Series):
        is_date_no_time = False

        # Use equality instead of isinstance because of inheritance
        if type(x[0]) == datetime.date:
            x = pd.to_datetime(x)
            is_date_no_time = True

        for f, mn, mx in zip(self._frequency, self._min, self._max):
            noise = self._rng.integers(mn, mx, size=x.shape)
            delta_fn = _FREQUENCY_TO_DELTA_FN.get(f, None)
            if delta_fn is None:
                raise ValueError(
                    "Frequency {} must be one of {}.".format(
                        f, list(_FREQUENCY_TO_DELTA_FN.keys())
                    )
                )
            x += delta_fn(noise)

        if is_date_no_time:
            return pd.Series(x).dt.date
        else:
            return x


def _check_minmax_arg(arg):
    """Checks that arg is an integer or a flat collection of integers."""
    if not isinstance(arg, (tuple, list)):
        if not isinstance(arg, int):
            raise ValueError
        return [arg]
    else:
        for a in arg:
            if not isinstance(a, int):
                raise ValueError
    return arg


def _check_freq_arg(arg):
    """Checks that arg in one of the frequency options."""
    if not isinstance(arg, (tuple, list)):
        if not isinstance(arg, str):
            raise ValueError
        return [arg]
    else:
        for a in arg:
            if not isinstance(a, str):
                raise ValueError
    return arg



from cape_privacy.utils import typecheck


def _make_args_and_types():
    string = "hi"
    integer = 4
    flt = 2.0
    lst = [string, integer, flt]
    tpl = (string, integer, float)
    args_list = (string, integer, flt, lst, tpl, None)
    types_list = (str, int, float, list, tuple, type(None))
    return args_list, types_list


def test_typecheck_args():
    args_list, types_list = _make_args_and_types()
    # check passing
    for a, t in zip(args_list, types_list):
        typecheck.check_arg(a, t)
    # check failure
    for a, t in zip(args_list, types_list[::-1]):
        try:
            typecheck.check_arg(a, t)
            raise AssertionError
        except ValueError:
            pass


def test_typecheck_more_types_passes():
    args_list, types_list = _make_args_and_types()
    args = args_list[:3]
    types = types_list[:3]
    for arg in args:
        typecheck.check_arg(arg, types)


def test_typecheck_more_types_fails():
    args_list, types_list = _make_args_and_types()
    arg = args_list[-1]
    types = types_list[:3]
    try:
        typecheck.check_arg(arg, types)
        raise AssertionError
    except ValueError:
        pass



from cape_privacy import pandas
from cape_privacy import spark
from cape_privacy.policy.policy import apply_policy
from cape_privacy.policy.policy import parse_policy

__all__ = ["apply_policy", "pandas", "parse_policy", "spark"]



import datetime

import numpy as np
import pandas as pd
import pandas.testing as pdt

from cape_privacy.pandas import dtypes
from cape_privacy.pandas.transformations import DatePerturbation
from cape_privacy.pandas.transformations import NumericPerturbation


def test_perturbation_float():
    transform = NumericPerturbation(dtype=dtypes.Float, min=-10, max=10, seed=1234)

    df = pd.DataFrame({"amount": range(5)})
    expected = pd.DataFrame(
        {"amount": [9.53399, -1.39608, 10.46492, -1.76615, 0.38194]}
    )

    df["amount"] = transform(df.amount)

    pdt.assert_frame_equal(df, expected)


def test_perturbation_int():
    transform = NumericPerturbation(dtype=dtypes.Integer, min=-10, max=10, seed=12345)

    df = pd.DataFrame({"amount": range(5)})
    expected = pd.DataFrame({"amount": [-5, -2, 7, 6, 2]})

    df["amount"] = transform(df.amount)

    pdt.assert_frame_equal(df, expected)


def test_perturbation_datetime():
    transform = DatePerturbation(frequency="DAY", min=-10, max=10, seed=1234)

    df = pd.DataFrame({"date": [np.datetime64("2018-10-15")]})
    expected = pd.DataFrame({"date": [np.datetime64("2018-10-24")]})

    df["date"] = transform(df.date)

    pdt.assert_frame_equal(df, expected)


def test_perturbation_date():
    transform = DatePerturbation(frequency="DAY", min=-10, max=10, seed=1234)

    df = pd.DataFrame({"date": [datetime.date(year=2018, month=10, day=15)]})
    expected = pd.DataFrame({"date": [datetime.date(year=2018, month=10, day=24)]})

    df["date"] = transform(df.date)

    pdt.assert_frame_equal(df, expected)


def test_perturbation_dat_mutliple_freq():
    transform = DatePerturbation(
        frequency=("DAY", "YEAR"), min=(-10, -5), max=(10, 5), seed=1234
    )

    df = pd.DataFrame({"date": [np.datetime64("2018-10-15")]})
    expected = pd.DataFrame({"date": [np.datetime64("2022-10-23")]})

    df["date"] = transform(df.date)

    pdt.assert_frame_equal(df, expected)
